"""
Defines the highest-level method for task handling through AWS SQS.
"""

import json
import logging
import os

import fire

from spacer import config
from spacer.messages import JobMsg, DataLocation
from spacer.tasks import \
    process_job


def env_job(): # pragma: no cover
    """ Runs a job defined in environmental variables. Setup to play
    nicely with AWS Batch.
    Expects JOB_MSG_LOC to contain the json-serialized DataLocation of JobMgs.
    Expects OUT_QUEUE to contain the name of the desired results queue.
    """

    job_msg_loc = os.getenv('JOB_MSG_LOC')
    if job_msg_loc is None:
        raise ValueError('JOB_MSG_LOC env. variable not set. '
                         'Can not process job.')

    logging.info(" Received job for ENV {}.".format(job_msg_loc))

    with config.log_entry_and_exit('job message location deserialization'):
        job_msg_loc = DataLocation.deserialize(json.loads(job_msg_loc))

    with config.log_entry_and_exit('job message initialization'):
        job_msg = JobMsg.load(job_msg_loc)

    try:
        job_return_msg = process_job(job_msg)
        job_return_msg_dict = job_return_msg.serialize()
    except Exception as e:
        # Handle deserialization errors directly in mailman.
        # All other errors are handled in "handle_message" function.
        job_return_msg_dict = {
            'original_job': job_msg.serialize(),
            'ok': False,
            'results': None,
            'error_message': 'Error deserializing message: ' + repr(e)
        }

    # Return
    out_queue_name = os.getenv('RES_SQS_QUEUE')
    if out_queue_name is not None:
        with config.log_entry_and_exit('writing results to {}'.format(
                out_queue_name)):
            conn = config.get_sqs_conn()
            out_queue = conn.get_queue(out_queue_name)
            m_out = out_queue.new_message(body=json.dumps(job_return_msg_dict))
            out_queue.write(m_out)

    out_msg_loc = os.getenv('RES_MSG_LOC')
    if out_msg_loc is not None:
        with config.log_entry_and_exit('out message location deserialization'):
            out_msg_loc = DataLocation.deserialize(json.loads(out_msg_loc))

        with config.log_entry_and_exit('writing res to {}'.format(
                out_msg_loc.key)):
            s3 = config.get_s3_conn()
            s3.Bucket(out_msg_loc.bucket_name).put_object(
                Key=out_msg_loc.key,
                Body=bytes(json.dumps(job_return_msg_dict).encode('UTF-8')))
    return 1


if __name__ == '__main__':
    fire.Fire()  # pragma: no cover



from typing import List, Tuple

from PIL import Image


def check_rowcols(rowcols: List[Tuple[int, int]],
                  image: Image):

    im_width, im_height = image.size

    for row, col in rowcols:
        assert(isinstance(row, int)), \
            "Rows must be integers. Given was: [{}]".format(row)

        assert (isinstance(col, int)), \
            "Columns must be integers. Given was: [{}]".format(col)

        assert (row >= 0), \
            "Rows must be non-negative. Given was: [{}]".format(row)

        assert (col >= 0), \
            "Columns must be non-negative. Given was: [{}]".format(col)

        assert row < im_height, \
            "Row {} outside image with nrows: {}".format(row, im_height)

        assert col < im_width, \
            "Column {} outside image with ncols: {}".format(col, im_width)



"""
Contains config and settings for the repo.
"""

import importlib
import json
import logging
import os
import time
import warnings
from contextlib import ContextDecorator
from typing import Tuple, Optional

import boto3
import botocore.exceptions
from PIL import Image

# Configure a simple logger that works with AWS cloudwatch
if len(logging.getLogger().handlers) > 0:
    # The Lambda environment pre-configures a handler logging to stderr.
    # If a handler is already configured,
    # `.basicConfig` does not execute. Thus we set the level directly.
    logging.getLogger().setLevel(logging.INFO)
else:
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')


def filter_warnings():
    """ Filters out some verified warnings. """

    # Per discussion in https://github.com/boto/boto3/issues/454,
    # the boto package is raising a lot of warnings that it shouldn't.
    warnings.filterwarnings("ignore", category=ResourceWarning,
                            message="unclosed.*<ssl.SSLSocket.*>")
    warnings.filterwarnings("ignore", category=ResourceWarning,
                            message="unclosed.*<_io.TextIOWrapper.*>")


def get_secret(key):
    """ Try to load settings from secrets.json file """
    secrets_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                '..', 'secrets.json')
    if not os.path.exists(secrets_path):  # pragma: no cover
        RuntimeWarning('secrets.json not found')
        return None
    else:
        try:
            with open(secrets_path) as fp:
                secrets = json.load(fp)
            return secrets[key]
        except Exception as err_:  # pragma: no cover
            RuntimeWarning(
                'Unable to parse secrets.json: {}'.format(repr(err_)))
            return None


def get_aws_credentials() -> Tuple[Optional[str], Optional[str]]:
    aws_key_id = os.getenv('SPACER_AWS_ACCESS_KEY_ID')
    aws_key_secret = os.getenv('SPACER_AWS_SECRET_ACCESS_KEY')

    if not aws_key_id:
        aws_key_id = get_secret('SPACER_AWS_ACCESS_KEY_ID')
    if not aws_key_secret:
        aws_key_secret = get_secret('SPACER_AWS_SECRET_ACCESS_KEY')

    return aws_key_id, aws_key_secret


def get_s3_conn():
    """
    Returns a boto s3 connection.
    - It first looks for credentials in the environmental vars.
    - If not found there it looks in secrets.json
    - If not found there it will default to credentials in ~/.aws/credentials
    """
    aws_key_id, aws_key_secret = get_aws_credentials()

    return boto3.resource('s3',
                          region_name="us-west-2",
                          aws_access_key_id=aws_key_id,
                          aws_secret_access_key=aws_key_secret)


def get_sqs_conn():
    """
    Returns a connection to SQS.
    - It first looks for credentials in the environmental vars.
    - If not found there it looks in secrets.json
    - If not found there it will default to credentials in ~/.aws/credentials
    """
    aws_key_id, aws_key_secret = get_aws_credentials()
    return boto3.resource('sqs',
                          region_name="us-west-2",
                          aws_access_key_id=aws_key_id,
                          aws_secret_access_key=aws_key_secret)


def get_local_model_path():
    local_model_path = os.getenv('SPACER_LOCAL_MODEL_PATH')
    if local_model_path is None:
        return get_secret('SPACER_LOCAL_MODEL_PATH')  # pragma: no cover
    return local_model_path


class log_entry_and_exit(ContextDecorator):
    def __init__(self, name):
        self.name = name
        self.start_time = None

    def __enter__(self):
        self.start_time = time.time()
        logging.info('Entering: %s', self.name)

    def __exit__(self, exc_type, exc, exc_tb):
        logging.info('Exiting: %s after %f seconds.', self.name,
                     time.time() - self.start_time)


LOCAL_MODEL_PATH = get_local_model_path()

HAS_LOCAL_MODEL_PATH = LOCAL_MODEL_PATH is not None and \
                       os.path.exists(LOCAL_MODEL_PATH)

TASKS = [
    'extract_features',
    'train_classifier',
    'classify_features',
    'classify_image'
]

FEATURE_EXTRACTOR_NAMES = [
    'dummy',
    'vgg16_coralnet_ver1',
    'efficientnet_b0_ver1'
]

MODEL_WEIGHTS_SHA = {
    'vgg16':
        'fb83781de0e207ded23bd42d7eb6e75c1e915a6fbef74120f72732984e227cca',
    'efficientnet-b0':
        'c3dc6d304179c6729c0a0b3d4e60c728bdcf0d82687deeba54af71827467204c',
}

CLASSIFIER_TYPES = [
    'LR',
    'MLP'
]

TRAINER_NAMES = [
    'minibatch'
]

MODELS_BUCKET = 'spacer-tools'
TEST_BUCKET = 'spacer-test'

STORAGE_TYPES = [
    's3',
    'filesystem',
    'memory',
    'url'
]

MAX_IMAGE_PIXELS = 10000 * 10000  # 100 mega pixels is max we allow.
MAX_POINTS_PER_IMAGE = 1000

LOCAL_FIXTURE_DIR = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'tests', 'fixtures')

# The train_classifier task require as least this many images.
MIN_TRAINIMAGES = 10

# Check access to select which tests to run.
HAS_CAFFE = importlib.util.find_spec("caffe") is not None


try:
    s3 = get_s3_conn()
    s3.meta.client.head_bucket(Bucket=TEST_BUCKET)
    HAS_S3_TEST_ACCESS = True
except (botocore.exceptions.ClientError,
        botocore.exceptions.NoCredentialsError):  # pragma: no cover
    logging.info("No connection to spacer-test bucket, "
                 "can't run remote tests")
    HAS_S3_TEST_ACCESS = False

try:
    s3 = get_s3_conn()
    s3.meta.client.head_bucket(Bucket=MODELS_BUCKET)
    HAS_S3_MODEL_ACCESS = True
except (botocore.exceptions.ClientError,
        botocore.exceptions.NoCredentialsError):  # pragma: no cover
    logging.info("No connection to spacer-tools bucket, "
                 "can't run remote tests")
    HAS_S3_MODEL_ACCESS = False

# Add margin to avoid warnings when running unit-test.
Image.MAX_IMAGE_PIXELS = MAX_IMAGE_PIXELS + 20000



"""
For possible future network extension
"""
from .efficientnet import EfficientNet


def get_model(model_type, model_name, num_classes, pretrained=False):
    if model_type == 'efficientnet':
        model = EfficientNet.from_pretrained(model_name=model_name,
                                             num_classes=num_classes)
    else:
        raise NotImplementedError(model_name)
    return model



"""
Defines storage ABC; implementations; and factory.
"""

import abc
import logging
import os
import pickle
import warnings
from functools import lru_cache
from io import BytesIO
from typing import Union, Tuple
from urllib.error import URLError

import botocore.exceptions
import wget
from PIL import Image
from sklearn.calibration import CalibratedClassifierCV

from spacer import config


class Storage(abc.ABC):  # pragma: no cover

    @abc.abstractmethod
    def store(self, key: str, stream: BytesIO) -> None:
        """ Stores a BytesIO stream """

    @abc.abstractmethod
    def load(self, key: str) -> BytesIO:
        """ Loads key to BytesIO stream """

    @abc.abstractmethod
    def delete(self, key: str) -> None:
        """ Deletes key if it exists """

    @abc.abstractmethod
    def exists(self, key: str) -> bool:
        """ Checks if key exists """


class URLStorage(Storage):
    """ Loads items from URLs. Does not support store operations. """

    def __init__(self):
        self.fs_storage = FileSystemStorage()

    def store(self, url: str, stream: BytesIO):
        raise TypeError('Store operation not supported for URL storage.')

    def load(self, url: str):
        tmp_path = wget.download(url)
        stream = self.fs_storage.load(tmp_path)
        self.fs_storage.delete(tmp_path)
        return stream

    def delete(self, url: str) -> None:
        raise TypeError('Delete operation not supported for URL storage.')

    def exists(self, url: str) -> bool:
        try:
            tmp_path = wget.download(url)
        except URLError:
            return False
        except ValueError:
            return False
        self.fs_storage.delete(tmp_path)
        return True


class S3Storage(Storage):
    """ Stores objects on AWS S3 """

    def __init__(self, bucketname: str):
        self.bucketname = bucketname

    def store(self, key: str, stream: BytesIO):
        s3 = config.get_s3_conn()
        s3.Bucket(self.bucketname).put_object(Body=stream, Key=key)

    def load(self, key: str):
        s3 = config.get_s3_conn()
        stream = BytesIO()
        s3.Object(self.bucketname, key).download_fileobj(stream)
        return stream

    def delete(self, key: str) -> None:
        s3 = config.get_s3_conn()
        s3.Object(self.bucketname, key).delete()

    def exists(self, key: str):
        s3 = config.get_s3_conn()
        try:
            s3.Object(self.bucketname, key).load()
            return True
        except botocore.exceptions.ClientError:
            return False


class FileSystemStorage(Storage):
    """ Stores objects on disk """

    def __init__(self):
        pass

    def store(self, key: str, stream: BytesIO):

        dirname = os.path.dirname(os.path.abspath(key))
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(key, 'wb') as f:
            f.write(stream.getbuffer())

    def load(self, key: str):
        with open(key, 'rb') as f:
            return BytesIO(f.read())

    def delete(self, path: str):
        os.remove(path)

    def exists(self, path: str):
        return os.path.exists(path)


class MemoryStorage(Storage):
    """ This stores objects in RAM. Useful for testing only. """

    def __init__(self):
        self.blobs = {}

    def store(self, key: str, stream: BytesIO):
        self.blobs[key] = stream.getvalue()

    def load(self, key: str):
        stream = BytesIO(self.blobs[key])
        return stream

    def delete(self, path: str):
        del self.blobs[path]

    def exists(self, path: str):
        return path in self.blobs


# This holds the global memory storage.
_memorystorage = None


def clear_memory_storage():
    """ Clears global memory storage"""
    global _memorystorage
    _memorystorage = None


def storage_factory(storage_type: str, bucketname: Union[str, None] = None):

    assert storage_type in config.STORAGE_TYPES

    if storage_type == 's3':
        return S3Storage(bucketname=bucketname)
    if storage_type == 'filesystem':
        return FileSystemStorage()
    if storage_type == 'memory':
        global _memorystorage
        if _memorystorage is None:
            _memorystorage = MemoryStorage()
        return _memorystorage
    if storage_type == 'url':
        return URLStorage()


def download_model(keyname: str) -> Tuple[str, bool]:
    """
    Utility method to download model to local cache.
    This is not part of the storage interface since it is a special case,
    and models need to be downloaded to a specific destination folder to be
    shared with host filesystem.
    """
    assert config.HAS_S3_MODEL_ACCESS, "Need access to model bucket."
    assert config.HAS_LOCAL_MODEL_PATH, "Model path not set or is invalid."
    destination = os.path.join(config.LOCAL_MODEL_PATH, keyname)

    with config.log_entry_and_exit('fetching of model to ' + destination):
        if not os.path.isfile(destination):
            with config.log_entry_and_exit('downloading from ' + keyname):
                s3 = config.get_s3_conn()
                s3.Bucket(config.MODELS_BUCKET).download_file(keyname,
                                                              destination)
                was_cashed = False
        else:
            # Already cached, no need to download
            was_cashed = True

    return destination, was_cashed


def store_image(loc: 'DataLocation', img: Image):
    storage = storage_factory(loc.storage_type, loc.bucket_name)
    with BytesIO() as stream:
        img.save(stream, 'JPEG')
        stream.seek(0)
        storage.store(loc.key, stream)


def load_image(loc: 'DataLocation'):
    storage = storage_factory(loc.storage_type, loc.bucket_name)
    return Image.open(storage.load(loc.key))


def store_classifier(loc: 'DataLocation', clf: CalibratedClassifierCV):
    storage = storage_factory(loc.storage_type, loc.bucket_name)
    storage.store(loc.key, BytesIO(pickle.dumps(clf, protocol=2)))


@lru_cache(maxsize=3)
def load_classifier(loc: 'DataLocation'):

    # This warning is due to the sklearn 0.17.1, 0.22.2 migration.
    warnings.filterwarnings('ignore', category=UserWarning,
                            message="Trying to unpickle.*")

    # This future warning also has to do with the unpickling of the
    # legacy model. It uses a to-be-deprecated import statement.
    warnings.filterwarnings('ignore', category=FutureWarning,
                            message="The sklearn.linear_model.*")

    def patch_legacy():
        """
        Upgrades models trained on scikit-learn 0.17.1 to 0.22.2
        Note: this in only tested for inference.
        """
        from sklearn.calibration import LabelEncoder

        logging.info("Patching legacy classifier.")
        assert len(clf.calibrated_classifiers_) == 1
        assert all(clf.classes_ == clf.calibrated_classifiers_[0].classes_)
        clf.calibrated_classifiers_[0].label_encoder_ = LabelEncoder()
        clf.calibrated_classifiers_[0].label_encoder_.fit(clf.classes_)

    storage = storage_factory(loc.storage_type, loc.bucket_name)
    clf = pickle.loads(storage.load(loc.key).getbuffer(), fix_imports=True,
                       encoding='latin1')

    if hasattr(clf, 'calibrated_classifiers_') and not \
            hasattr(clf.calibrated_classifiers_[0], 'label_encoder'):
        patch_legacy()

    return clf



"""
Defines train-classifier ABC; implementations; and factory.
"""

import abc
import time
from typing import Tuple, List

from sklearn.calibration import CalibratedClassifierCV

from spacer import config
from spacer.data_classes import ImageLabels, ValResults
from spacer.messages import TrainClassifierReturnMsg, DataLocation
from spacer.train_utils import train, evaluate_classifier, calc_acc


class ClassifierTrainer(abc.ABC):  # pragma: no cover

    @abc.abstractmethod
    def __call__(self,
                 train_labels: ImageLabels,
                 val_labels: ImageLabels,
                 nbr_epochs: int,
                 pc_models: List[CalibratedClassifierCV],
                 feature_loc: DataLocation,
                 clf_type: str) \
            -> Tuple[CalibratedClassifierCV,
                     ValResults,
                     TrainClassifierReturnMsg]:
        pass


class MiniBatchTrainer(ClassifierTrainer):
    """
    This is the default trainer. It uses mini-batches of data
    to train the classifier
    """

    def __call__(self,
                 train_labels,
                 val_labels,
                 nbr_epochs,
                 pc_models,
                 feature_loc,
                 clf_type):

        assert clf_type in config.CLASSIFIER_TYPES
        # Train.
        t0 = time.time()
        clf, ref_accs = train(train_labels, feature_loc, nbr_epochs, clf_type)
        classes = clf.classes_.tolist()

        # Evaluate new classifier on validation set.
        val_gts, val_ests, val_scores = evaluate_classifier(
            clf, val_labels, classes, feature_loc)

        # Evaluate previous classifiers on validation set.
        pc_accs = []
        for pc_model in pc_models:
            pc_gts, pc_ests, _ = evaluate_classifier(pc_model, val_labels,
                                                     classes, feature_loc)
            pc_accs.append(calc_acc(pc_gts, pc_ests))

        return \
            clf, \
            ValResults(
                scores=val_scores,
                gt=[classes.index(member) for member in val_gts],
                est=[classes.index(member) for member in val_ests],
                classes=classes
            ), \
            TrainClassifierReturnMsg(
                acc=calc_acc(val_gts, val_ests),
                pc_accs=pc_accs,
                ref_accs=ref_accs,
                runtime=time.time() - t0
            )


def trainer_factory(trainer_name: str) -> ClassifierTrainer:
    """ There is only one type of Trainer, so this factory is trivial. """
    assert trainer_name in config.TRAINER_NAMES
    if trainer_name == 'minibatch':
        return MiniBatchTrainer()



"""
Defines data-classes for input and output types.
Each data-class can serialize itself to a structure of JSON-friendly
python-native data-structures such that it can be stored.
"""
import json
from abc import ABC, abstractmethod
from io import BytesIO
from pprint import pformat
from typing import Dict, List, Tuple, Set, Optional, Union

import numpy as np

from spacer.storage import storage_factory


class DataClass(ABC):  # pragma: no cover

    @classmethod
    def load(cls, loc: 'DataLocation'):
        storage = storage_factory(loc.storage_type, loc.bucket_name)
        return cls.deserialize(json.loads(
            storage.load(loc.key).getvalue().decode('utf-8')))

    def store(self, loc: 'DataLocation'):
        storage = storage_factory(loc.storage_type, loc.bucket_name)
        storage.store(loc.key, BytesIO(
            json.dumps(self.serialize()).encode('utf-8')))

    @classmethod
    @abstractmethod
    def example(cls):
        """
        Instantiate an example member of the class.
        Useful for testing, tutorials, etc.
        """

    @classmethod
    def deserialize(cls, data: Dict) -> 'DataClass':
        """
        Initializes a class instance from input data.
        Input should be a dictionary containing native python data structures.
        """
        return cls(**data)

    def serialize(self) -> Dict:
        """
        Serialized the class content to native data-types, dicts, lists, etc,
        such that it it compatible with json.dumps and json.loads.
        """
        return self.__dict__

    def __repr__(self):
        """ Give a (sort of) nicely formatted string repr. of class. """
        return pformat(vars(self))

    def __eq__(self, other):
        """
        This only works for the simple classes, more complicated
        need to overwrite this method
        """
        sd = self.__dict__
        od = other.__dict__
        return sd.keys() == od.keys() and all([sd[key] == od[key]
                                               for key in sd])


class ImageLabels(DataClass):
    """ Contains row, col, label information for an image. """

    def __init__(self,
                 # Data maps a feature key (or file path) to a List of
                 # (row, col, label).
                 data: Dict[str, List[Tuple[int, int, int]]]):
        self.data = data

    @classmethod
    def example(cls):
        return ImageLabels({
            'img1.features': [(100, 200, 3), (101, 200, 2)],
            'img2.features': [(100, 202, 3), (101, 200, 3)],
            'img3.features': [(100, 202, 3), (101, 200, 3)],
            'img4.features': [(100, 202, 3), (101, 200, 3)],
        })

    @classmethod
    def deserialize(cls, data: Dict) -> 'ImageLabels':
        """ Custom deserializer required to convert back to tuples. """
        return ImageLabels(
            data={key: [tuple(entry) for entry in value] for
                  key, value in data['data'].items()})

    @property
    def image_keys(self):
        return list(self.data.keys())

    @property
    def samples_per_image(self):
        return len(next(iter(self.data.values())))

    def unique_classes(self, key_list: List[str]) -> Set[int]:
        """ Returns the set of all unique classes in the key_list subset. """
        labels = set()
        for im_key in key_list:
            labels |= {label for (row, col, label) in self.data[im_key]}
        return labels

    def __len__(self):
        return len(self.data)


class PointFeatures(DataClass):
    """ Contains row, col, feature-vector for a single point. """

    def __init__(self,
                 row: Optional[int],  # Row where feature was extracted
                 col: Optional[int],  # Column where feature was extracted
                 data: np.array,  # Feature vector with half precision floats.
                 ):
        self.row = row
        self.col = col
        self.data = np.array(data, dtype=np.half)

    @classmethod
    def example(cls):
        return cls(
            row=100,
            col=100,
            data=np.array([1.1, 1.3, 1.12], dtype=np.half)
        )

    def __eq__(self, other):
        return self.row == other.row and self.col == other.col and \
               np.allclose(self.data, other.data)


class ImageFeatures(DataClass):
    """ Contains row, col, feature-vectors for an image. """

    def __init__(self,
                 # List of features for all points in image
                 point_features: List[PointFeatures],
                 # Legacy feature did not store row and column locations.
                 valid_rowcol: bool,
                 # Dimensionality of the feature vectors.
                 feature_dim: int,
                 # Number of points in this image.
                 npoints: int,
                 ):

        assert len(point_features) == npoints
        assert len(point_features) > 0
        assert len(point_features[0].data) == feature_dim

        self.point_features = point_features
        self.valid_rowcol = valid_rowcol
        self.feature_dim = feature_dim
        self.npoints = npoints

        if self.valid_rowcol:
            # Store a row_col hash for quick retrieval based on (row, col)
            self._rchash = {(pf.row, pf.col): enum for
                            enum, pf in enumerate(self.point_features)}

    def __getitem__(self, rowcol: Tuple[int, int]) -> List[float]:
        """ Returns features at (row, col) location. """
        if not self.valid_rowcol:
            raise ValueError('Method not supported for legacy features')
        return self.point_features[self._rchash[rowcol]].data

    def get_array(self, rowcol: Tuple[int, int]) -> np.array:
        """
        Similar to __getitem__ but returns a numpy array formatted
        correctly for classification by a classifier of type
        sklearn.calibration.CalibratedClassifierCV
        """
        return np.array(self[rowcol]).reshape(1, -1)

    @classmethod
    def example(cls):
        pf1 = PointFeatures(row=100, col=100, data=[1.1, 1.3, 1.12])
        pf2 = PointFeatures(row=120, col=110, data=[1.9, 1.3, 1.12])
        return cls(
            point_features=[pf1, pf2],
            valid_rowcol=True,
            feature_dim=len(pf1.data),
            npoints=2
        )

    @classmethod
    def deserialize(cls, data: Union[Dict, List]) -> 'ImageFeatures':

        assert isinstance(data, List), \
            "Deserialize only supported for legacy format"
        # Legacy feature were stored as a list of list
        # without row and column information.
        assert len(data) > 0, "Empty features file."
        return ImageFeatures(
            point_features=[PointFeatures(row=None,
                                          col=None,
                                          data=entry) for entry in data],
            valid_rowcol=False,
            feature_dim=len(data[0]),
            npoints=len(data)
        )

    def serialize(self):
        raise NotImplementedError('Use .store() and .load() methods instead.')

    def __eq__(self, other):
        return all([a == b for a, b in zip(self.point_features,
                                           other.point_features)]) and \
               self.valid_rowcol == other.valid_rowcol and \
               self.feature_dim == other.feature_dim and \
               self.npoints == other.npoints

    @classmethod
    def make_random(cls,
                    point_labels: List[int],
                    feature_dim: int):
        pfs = [PointFeatures(row=itt,
                             col=itt,
                             data=list(label + np.random.randn(feature_dim)))
               for itt, label in enumerate(point_labels)]

        return ImageFeatures(point_features=pfs,
                             valid_rowcol=True,
                             feature_dim=feature_dim,
                             npoints=len(point_labels))

    @classmethod
    def load(cls, loc: 'DataLocation'):

        storage = storage_factory(loc.storage_type, loc.bucket_name)
        stream = storage.load(loc.key)
        stream.seek(0)
        try:
            data = np.load(stream)
            valid_rowcol = data['meta'][0]
            if valid_rowcol:
                return ImageFeatures(
                    point_features=[PointFeatures(
                        row=int(row),
                        col=int(col),
                        data=feat) for row, col, feat in zip(data['rows'],
                                                             data['cols'],
                                                             data['feat'])],
                    valid_rowcol=data['meta'][0],
                    npoints=data['meta'][1],
                    feature_dim=data['meta'][2]
                )
            else:
                return cls.deserialize(data['feat'].tolist())

        except ValueError:
            "We used to store these as a JSON file with a list of features."
            data = json.loads(stream.getvalue().decode('utf-8'))
            return cls.deserialize(data)

    def store(self, loc: 'DataLocation'):
        storage = storage_factory(loc.storage_type, loc.bucket_name)
        if self.valid_rowcol:
            rows = np.array([p.row for p in self.point_features], dtype=np.uint16)
            cols = np.array([p.col for p in self.point_features], dtype=np.uint16)
        else:
            rows = np.array([])
            cols = np.array([])
        feat = np.array([p.data for p in self.point_features], dtype=np.half)
        meta = np.array([self.valid_rowcol, self.npoints, self.feature_dim])
        output = BytesIO()
        np.savez_compressed(output, meta=meta, rows=rows, cols=cols, feat=feat)
        output.seek(0)
        storage.store(loc.key, output)


class ValResults(DataClass):
    """ Defines the validation results data class. Note that the gt and est
    lists points to the index into the classes list."""

    def __init__(self,
                 scores: List[float],
                 gt: List[int],  # Using singular for backwards compatibility.
                 est: List[int],  # Using singular for backwards compatibility.
                 classes: List[int]):

        assert len(gt) == len(est)
        assert len(gt) == len(scores)
        assert max(gt) < len(classes)
        assert max(est) < len(classes)
        assert min(gt) >= 0
        assert min(est) >= 0

        self.scores = scores
        self.gt = gt
        self.est = est
        self.classes = classes

    @classmethod
    def example(cls):
        return cls(scores=[.9, .8, .7],
                   gt=[0, 1, 0],
                   est=[0, 1, 1],
                   classes=[121, 1222])

    @classmethod
    def deserialize(cls, data: Dict) -> 'ValResults':
        """ Redefined here to help the Typing module. """
        return ValResults(**data)



"""
Defines the highest level methods for completing tasks.
"""
import logging
import time
import traceback

from spacer import config
from spacer.data_classes import ImageFeatures
from spacer.extract_features import feature_extractor_factory
from spacer.messages import \
    ExtractFeaturesMsg, \
    ExtractFeaturesReturnMsg, \
    TrainClassifierMsg, \
    TrainClassifierReturnMsg, \
    ClassifyFeaturesMsg, \
    ClassifyImageMsg, \
    ClassifyReturnMsg, JobMsg, JobReturnMsg
from spacer.storage import load_image, load_classifier, store_classifier
from spacer.task_utils import check_rowcols
from spacer.train_classifier import trainer_factory


def extract_features(msg: ExtractFeaturesMsg) -> ExtractFeaturesReturnMsg:

    extractor = feature_extractor_factory(msg.feature_extractor_name)
    img = load_image(msg.image_loc)

    assert img.width * img.height <= config.MAX_IMAGE_PIXELS, \
        "Image ({}, {}) with {} pixels too large. (max: {})".format(
            img.width, img.height, img.width * img.height,
            config.MAX_IMAGE_PIXELS)
    assert len(msg.rowcols) <= config.MAX_POINTS_PER_IMAGE, \
        "Too many rowcol locations ({}). Max {} allowed".format(
            len(msg.rowcols), config.MAX_POINTS_PER_IMAGE
        )

    check_rowcols(msg.rowcols, img)

    with config.log_entry_and_exit('actual extraction'):
        features, return_msg = extractor(img, msg.rowcols)

    with config.log_entry_and_exit('storing features'):
        features.store(msg.feature_loc)

    return return_msg


def train_classifier(msg: TrainClassifierMsg) -> TrainClassifierReturnMsg:
    trainer = trainer_factory(msg.trainer_name)

    # Do the actual training
    with config.log_entry_and_exit('actual training'):
        clf, val_results, return_message = trainer(
            msg.train_labels,
            msg.val_labels,
            msg.nbr_epochs,
            [load_classifier(loc) for loc in msg.previous_model_locs],
            msg.features_loc,
            msg.clf_type
        )

    with config.log_entry_and_exit('storing classifier and val res'):
        store_classifier(msg.model_loc, clf)
        val_results.store(msg.valresult_loc)

    return return_message


def classify_features(msg: ClassifyFeaturesMsg) -> ClassifyReturnMsg:

    t0 = time.time()
    features = ImageFeatures.load(msg.feature_loc)

    clf = load_classifier(msg.classifier_loc)

    scores = [(pf.row, pf.col,
               clf.predict_proba(pf.data.reshape(1, -1)).tolist()[0]) for
              pf in features.point_features]

    # Return
    return ClassifyReturnMsg(
        runtime=time.time() - t0,
        scores=scores,
        classes=clf.classes_.tolist(),
        valid_rowcol=features.valid_rowcol)


def classify_image(msg: ClassifyImageMsg) -> ClassifyReturnMsg:

    t0 = time.time()

    # Download image
    img = load_image(msg.image_loc)
    check_rowcols(msg.rowcols, img)

    # Extract features
    extractor = feature_extractor_factory(msg.feature_extractor_name)
    features, _ = extractor(img, msg.rowcols)

    # Classify
    clf = load_classifier(msg.classifier_loc)
    scores = [(row, col, clf.predict_proba(features.get_array((row, col))).
               tolist()) for row, col in msg.rowcols]

    # Return
    return ClassifyReturnMsg(
        runtime=time.time() - t0,
        scores=scores,
        classes=list(clf.classes_),
        valid_rowcol=True)


def process_job(job_msg: JobMsg) -> JobReturnMsg:

    run = {
        'extract_features': extract_features,
        'train_classifier': train_classifier,
        'classify_features': classify_features,
        'classify_image': classify_image,
    }

    assert isinstance(job_msg, JobMsg)
    assert job_msg.task_name in config.TASKS

    results = []
    for task in job_msg.tasks:
        try:
            with config.log_entry_and_exit('{} [{}]'.format(
                    job_msg.task_name, task.job_token)):
                results.append(run[job_msg.task_name](task))
        except Exception:
            logging.error('Error executing job {}: {}'.format(
                task.job_token, traceback.format_exc()))
            return_msg = JobReturnMsg(
                original_job=job_msg,
                ok=False,
                results=None,
                error_message=traceback.format_exc()
            )
            return return_msg

    return_msg = JobReturnMsg(
        original_job=job_msg,
        ok=True,
        results=results,
        error_message=None
    )

    return return_msg



"""
Utility methods for training classifiers.
"""

import logging
import random
import string
from typing import Tuple, List

import numpy as np
from sklearn.calibration import CalibratedClassifierCV
from sklearn.linear_model import SGDClassifier
from sklearn.neural_network import MLPClassifier

from spacer import config
from spacer.data_classes import ImageLabels, ImageFeatures
from spacer.messages import DataLocation


def train(labels: ImageLabels,
          feature_loc: DataLocation,
          nbr_epochs: int,
          clf_type: str) -> Tuple[CalibratedClassifierCV, List[float]]:

    if len(labels) < config.MIN_TRAINIMAGES:
        raise ValueError('Not enough training samples.')

    # Calculate max nbr images to keep in memory (for 5000 samples total).
    max_imgs_in_memory = 5000 // labels.samples_per_image

    # Make train and ref split.
    # Reference set is here a hold-out part of the train-data portion.
    # Purpose of reference set is to
    # 1) know accuracy per epoch
    # 2) calibrate classifier output scores.
    # We call it "reference" to disambiguate from the validation set.
    ref_set = labels.image_keys[::10]
    np.random.shuffle(ref_set)
    ref_set = ref_set[:max_imgs_in_memory]  # Enforce memory limit.
    train_set = list(set(labels.image_keys) - set(ref_set))
    logging.info("Trainset: {}, valset: {} images".
                 format(len(train_set), len(ref_set)))

    # Figure out # images per batch and batches per epoch.
    images_per_batch, batches_per_epoch = \
        calc_batch_size(max_imgs_in_memory, len(train_set))
    logging.info("Using {} images per mini-batch and {} mini-batches per "
                 "epoch".format(images_per_batch, batches_per_epoch))

    # Identify classes common to both train and val.
    # This will be our labelset for the training.
    trainclasses = labels.unique_classes(train_set)
    refclasses = labels.unique_classes(ref_set)
    classes = list(trainclasses.intersection(refclasses))
    logging.info("Trainset: {}, valset: {}, common: {} labels".format(
        len(trainclasses), len(refclasses), len(classes)))
    if len(classes) == 1:
        raise ValueError('Not enough classes to do training (only 1)')

    # Load reference data (must hold in memory for the calibration)
    with config.log_entry_and_exit("loading of reference data"):
        refx, refy = load_batch_data(labels, ref_set, classes, feature_loc)

    # Initialize classifier and ref set accuracy list
    with config.log_entry_and_exit("training using " + clf_type):
        if clf_type == 'MLP':
            if len(train_set) * labels.samples_per_image >= 50000:
                hls, lr = (200, 100), 1e-4
            else:
                hls, lr = (100,), 1e-3
            clf = MLPClassifier(hidden_layer_sizes=hls, learning_rate_init=lr)
        else:
            clf = SGDClassifier(loss='log', average=True, random_state=0)
        ref_acc = []
        for epoch in range(nbr_epochs):
            np.random.shuffle(train_set)
            mini_batches = chunkify(train_set, batches_per_epoch)
            for mb in mini_batches:
                x, y = load_batch_data(labels, mb, classes, feature_loc)
                clf.partial_fit(x, y, classes=classes)
            ref_acc.append(calc_acc(refy, clf.predict(refx)))
            logging.info("Epoch {}, acc: {}".format(epoch, ref_acc[-1]))

    with config.log_entry_and_exit("calibration"):
        clf_calibrated = CalibratedClassifierCV(clf, cv="prefit")
        clf_calibrated.fit(refx, refy)

    return clf_calibrated, ref_acc


def evaluate_classifier(clf: CalibratedClassifierCV,
                        labels: ImageLabels,
                        classes: List[int],
                        feature_loc: DataLocation) -> Tuple[List, List, List]:
    """ Evaluates classifier on data """
    scores, gts, ests = [], [], []
    for imkey in labels.image_keys:
        x, y = load_image_data(labels, imkey, classes, feature_loc)
        if len(x) > 0:
            scores.extend(clf.predict_proba(x).max(axis=1).tolist())
            ests.extend(clf.predict(x))
            gts.extend(y)

    if len(gts) == 0:
        raise ValueError('Not enough data in validation set.')

    return gts, ests, scores


def chunkify(lst: List,
             nbr_chunks: int) -> List:
    return [lst[i::nbr_chunks] for i in range(nbr_chunks)]


def calc_batch_size(max_imgs_in_memory: int,
                    train_set_size: int) -> Tuple[int, int]:
    images_per_batch = min(max_imgs_in_memory, train_set_size)
    batches_per_epoch = int(np.ceil(train_set_size / images_per_batch))
    return images_per_batch, batches_per_epoch


def load_image_data(labels: ImageLabels,
                    imkey: str,
                    classes: List[int],
                    feature_loc: DataLocation) \
        -> Tuple[List[List[float]], List[int]]:
    """
    Loads features and labels for image and matches feature with labels.
    """

    # Load features for this image.
    feature_loc.key = imkey  # Set the relevant key here.
    image_features = ImageFeatures.load(feature_loc)

    # Load row, col, labels for this image.
    image_labels = labels.data[imkey]

    # Sanity check
    assert len(image_labels) == len(image_features.point_features), \
        "number of extracted features doesn't match the number of labels!"

    x, y = [], []
    if image_features.valid_rowcol:
        for row, col, label in image_labels:
            if label not in classes:
                # Remove samples for which the label is not in classes.
                continue
            x.append(image_features[(row, col)])
            y.append(label)

    else:
        # For legacy features, we didn't store the row, col information.
        # Instead rely on ordering.
        for (_, _, label), point_feature in zip(image_labels,
                                                image_features.point_features):
            if label not in classes:
                continue
            x.append(point_feature.data)
            y.append(label)

    return x, y


def load_batch_data(labels: ImageLabels,
                    imkeylist: List[str],
                    classes: List[int],
                    feature_loc: DataLocation) \
        -> Tuple[List[List[float]], List[int]]:
    """ Loads features and labels and match them together. """
    x, y = [], []
    for imkey in imkeylist:
        x_, y_ = load_image_data(labels, imkey, classes, feature_loc)
        x.extend(x_)
        y.extend(y_)
    return x, y


def calc_acc(gt: List[int], est: List[int]) -> float:
    """
    Calculate the accuracy of (agreement between) two integer valued list.
    """
    if len(gt) == 0 or len(est) == 0:
        raise TypeError('Inputs can not be empty')

    if not len(gt) == len(est):
        raise ValueError('Input gt and est must have the same length')

    for g in gt:
        if not int(g) == g:
            raise TypeError('Input gt must be an array of ints')

    for e in est:
        if not int(e) == e:
            raise TypeError('Input est must be an array of ints')

    return float(sum([(g == e) for (g, e) in zip(gt, est)])) / len(gt)


def make_random_data(im_count: int,
                     class_list: List[int],
                     points_per_image: int,
                     feature_dim: int,
                     feature_loc: DataLocation) -> ImageLabels:
    """
    Utility method for testing that generates an ImageLabels instance
    complete with stored ImageFeatures.
    """
    labels = ImageLabels(data={})
    for _ in range(im_count):

        # Generate random features (using labels to draw from a Gaussian).
        point_labels = np.random.choice(class_list, points_per_image).tolist()

        # Make sure all classes are present
        point_labels[:len(class_list)] = class_list
        feats = ImageFeatures.make_random(point_labels, feature_dim)

        # Generate a random string as imkey.
        imkey = ''.join(random.choice(string.ascii_uppercase + string.digits)
                        for _ in range(20))

        # Store
        feature_loc.key = imkey
        feats.store(feature_loc)
        labels.data[imkey] = [
            (pf.row, pf.col, pl) for pf, pl in
            zip(feats.point_features, point_labels)
        ]
    return labels



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyspacer",  # Replace with your own username
    version="0.2.9",
    author="Oscar Beijbom",
    author_email="oscar.beijbom@gmail.com",
    description="Spatial image analysis with caffe and pytorch backends.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/beijbom/pyspacer",
    packages=setuptools.find_packages(exclude=['scripts',
                                               'scripts.*',
                                               'scripts.aws.*',
                                               'scripts.regression.*',
                                               'spacer.tests',
                                               'spacer.tests.*']),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[
        'wget==3.2',
        'tqdm==4.43.0',
        'Pillow>=4.2.0',
        'numpy>=1.17.5',
        'scikit-learn==0.22.1',
        'scikit-image==0.15.0',
        'torch==1.4.0',
        'torchvision==0.5.0',
        'boto3',
        'botocore'
    ]
)



"""
This file contains a set of caffe utility functions copied into this repo for
simplicity. Since support for Caffe will be deprecate,
these are only lightly cleaned up from their original state.
"""

from copy import copy
from functools import lru_cache
from typing import List, Any, Tuple

import caffe
import hashlib
import numpy as np

from spacer import config


class Transformer:
    """
    Transformer is a class for preprocessing and deprocessing images
    according to the vgg16 pre-processing paradigm.
    (scaling and mean subtraction.).
    """

    def __init__(self, mean: Tuple = (0, 0, 0)) -> None:
        self.mean = np.array(mean, dtype=np.float32)
        self.scale = 1.0

    def preprocess(self, im: np.ndarray) -> np.ndarray:
        """
        preprocess() emulate the pre-processing occurring
        in the vgg16 caffe prototxt.
        :param im: numpy array.
        :return: normalized numpy array.
        """
        im = np.float32(im)
        im = im[:, :, ::-1]  # change to BGR
        im -= self.mean
        im *= self.scale
        im = im.transpose((2, 0, 1))

        return im

    def deprocess(self, im: np.ndarray) -> np.ndarray:
        """
        inverse of preprocess().
        :param im: normalized numpy array.
        :return: original image.
        """
        im = im.transpose((1, 2, 0))
        im /= self.scale
        im += self.mean
        im = im[:, :, ::-1]  # change to RGB

        return np.uint8(im)


def classify_from_imlist(im_list: List,
                         net: Any,
                         transformer: Transformer,
                         batch_size: int,
                         scorelayer: str = 'score',
                         startlayer: str = 'conv1_1') -> List:
    """
    classify_from_imlist classifies a list of images and returns
    estimated labels and scores.
    Only support classification nets (not FCNs).
    :param im_list: list of images to classify (each stored as a numpy array).
    :param net: caffe net object.
    :param transformer: transformer object as defined above.
    :param batch_size: batch size for the net.
    :param scorelayer: name of the score (the last conv) layer.
    :param startlayer: name of first convolutional layer.
    :return: features list.
    """
    with config.log_entry_and_exit('forward pass through net'):
        scorelist = []
        for b in range(len(im_list) // batch_size + 1):
            for i in range(batch_size):
                pos = b * batch_size + i
                if pos < len(im_list):
                    net.blobs['data'].data[i, :, :, :] = \
                        transformer.preprocess(im_list[pos])
            net.forward(start=startlayer)
            scorelist.extend(list(copy(net.blobs[scorelayer].data).
                                  astype(np.float)))

        scorelist = scorelist[:len(im_list)]

    return scorelist


@lru_cache(maxsize=1)
def load_net(modeldef_path: str,
             modelweighs_path: str) -> Any:
    """
    load pretrained net.
    :param modeldef_path: model path.
    :param modelweighs_path: pretrained weights path.
    :return: pretrained model.
    """
    # To verify that the correct weight is loaded
    with config.log_entry_and_exit("load net, check SHA"):
        with open(modelweighs_path, 'rb') as fp:
            sha256 = hashlib.sha256(fp.read()).hexdigest()
        assert sha256 == config.MODEL_WEIGHTS_SHA['vgg16']

    return caffe.Net(modeldef_path, modelweighs_path, caffe.TEST)


def classify_from_patchlist(patchlist: List,
                            pyparams: dict,
                            modeldef_path: str,
                            modelweighs_path: str,
                            scorelayer: str = 'score',
                            startlayer: str = 'conv1_1') -> List:
    """
    extract features of a list of patches
    :param patchlist: a list of patches (cropped images).
    :param pyparams: a set of parameters.
    :param modeldef_path: model path.
    :param modelweighs_path: pretrained weights path.
    :param scorelayer: name of the score (the last conv) layer.
    :param startlayer: name of first convolutional layer.
    :return: a list of features
    """
    # Setup caffe
    caffe.set_mode_cpu()
    net = load_net(modeldef_path, modelweighs_path)

    # Classify
    transformer = Transformer(pyparams['im_mean'])
    scorelist = classify_from_imlist(
        patchlist, net, transformer, pyparams['batch_size'],
        scorelayer=scorelayer, startlayer=startlayer
    )

    return scorelist
